#include <iostream>
#include <list>
#include <vector>

#include "llvm/Support/Host.h"
#include "llvm/Support/MemoryBuffer.h"
#include "llvm/ADT/IntrusiveRefCntPtr.h"
#include "llvm/ADT/OwningPtr.h"
#include "llvm/ADT/StringRef.h"
#include "llvm/Support/raw_ostream.h"
#include "clang/Basic/DiagnosticOptions.h"
#include "clang/Frontend/TextDiagnosticPrinter.h"
#include "clang/Frontend/CompilerInstance.h"
#include "clang/Frontend/CompilerInvocation.h"
#include "clang/Frontend/FrontendOptions.h"
#include "clang/Frontend/LangStandard.h"
#include "clang/Basic/TargetOptions.h"
#include "clang/Basic/TargetInfo.h"
#include "clang/Basic/FileManager.h"
#include "clang/Basic/SourceManager.h"
#include "clang/Basic/LangOptions.h"
#include "clang/Lex/Preprocessor.h"
#include "clang/Lex/HeaderSearch.h"
#include "clang/Lex/DirectoryLookup.h"
#include "clang/Lex/PreprocessingRecord.h"
#include "clang/AST/ASTContext.h"
#include "clang/Rewrite/Core/Rewriter.h"
#include "clang/Rewrite/Frontend/Rewriters.h"
#include "clang/Basic/Diagnostic.h"

#include "clang/AST/RecursiveASTVisitor.h"
#include "clang/AST/ASTConsumer.h"
#include "clang/Parse/ParseAST.h"


using namespace clang;
using namespace llvm;
using namespace llvm::cl;
using namespace std;

class MyVisitor : public RecursiveASTVisitor<MyVisitor> {
public:
  explicit MyVisitor(Rewriter &R) : TheRewriter(R) {} 
  
  int poolIndex = 1;  
  bool isAddedPoolHeader = false;

  bool VisitForStmt(ForStmt *forStmt) 
  {
    std::vector<std::string> varNames; 
    std::vector<DeclRefExpr*> mallocCalls;
    std::vector<Expr*> sizesForMalloc;
    std::vector<ForStmt*> forStmts; 
    Stmt *forBody = forStmt->getBody();  
    for (StmtIterator b = forBody->child_begin(), e = forBody->child_end(); b != e; ++b)
    {
       if (isa<BinaryOperator>(*b))
       {
         //search for: "varName = (type*)malloc(sizeofsth);"
         BinaryOperator *bin = cast<BinaryOperator>(*b);
         Expr *right =  bin->getRHS(); 
         Expr *left = bin->getLHS();
         if (isa<CStyleCastExpr>(right))
		 {
		   CStyleCastExpr *cStyleCastExpr = cast<CStyleCastExpr>(right);
		   Expr *subCast = cStyleCastExpr -> getSubExpr();
  		   if (isa<CallExpr>(subCast))
           {
             CallExpr *call = cast<CallExpr>(subCast);
             Expr *callee = call->getCallee();
             if (isa<ImplicitCastExpr>(callee))
			 {
               ImplicitCastExpr *impl = cast<ImplicitCastExpr>(callee);
               Expr *subImpl = impl->getSubExpr();
               if (isa<DeclRefExpr>(subImpl))
			   {
				 DeclRefExpr *decl = cast<DeclRefExpr>(subImpl);
				 if (decl->getNameInfo().getAsString() == "malloc")
				 {
				   //add var name to the varNames vector
                   varNames.push_back(GetNameOfVariable(left));
                   //add call->getArg(0); to the sizesForMalloc vector
                   sizesForMalloc.push_back(call->getArg(0));
                   //add decl to the mallocCalls vector
                   mallocCalls.push_back(decl);
                   //add link to the current for statement
                   forStmts.push_back(forStmt);
				 }
			   }
			 }	
		   }
 		 }  
       }
       //search for "type* varName = (type*)malloc(sizeofsth);"
       else if (isa<DeclStmt>(*b))
       {
       		std::cout << "DECL in for\n";
       		DeclStmt *declStmt = cast<DeclStmt>(*b);
       		Decl *subDecl = declStmt->getSingleDecl();
       		if (isa<VarDecl>(subDecl))
       		{
       			VarDecl *varDecl = cast<VarDecl>(subDecl);
       			Expr *subVarDecl = varDecl->getInit();
       			if (isa<CStyleCastExpr>(subVarDecl))
       			{
       				CStyleCastExpr *cStyleCastExpr = cast<CStyleCastExpr>(subVarDecl);
       				Expr *subCStyleCastExpr = cStyleCastExpr -> getSubExpr();
       				if (isa<CallExpr>(subCStyleCastExpr))
       				{
       					CallExpr *callExpr = cast<CallExpr>(subCStyleCastExpr);
       					Expr *callee = callExpr->getCallee();
             			if (isa<ImplicitCastExpr>(callee))
			 			{
               				ImplicitCastExpr *impl = cast<ImplicitCastExpr>(callee);
               				Expr *subImpl = impl->getSubExpr();
               				if (isa<DeclRefExpr>(subImpl))
			   				{
				 				DeclRefExpr *decl = cast<DeclRefExpr>(subImpl);
				 				if (decl->getNameInfo().getAsString() == "malloc")
				 				{
				 					//add var name to the varNames vector
                   					varNames.push_back(varDecl->getNameAsString());
                   					//add call->getArg(0); to the sizesForMalloc vector
                   					sizesForMalloc.push_back(callExpr->getArg(0));
                   					//add decl to the mallocCalls vector
                   					mallocCalls.push_back(decl);
                   					//add link to the current for statement
                   					forStmts.push_back(forStmt);
				 				}
				 			}
				 		}
       				}
       			} 
       		}
       }
       //search for "free(a)"
	   else if (isa<CallExpr>(*b))
	   {
  	     CallExpr *call = cast<CallExpr>(*b);
		 Expr *callee = call->getCallee();
         if (isa<ImplicitCastExpr>(callee))
         {
           ImplicitCastExpr *impl = cast<ImplicitCastExpr>(callee);
           Expr *subImpl = impl -> getSubExpr();
		   if (isa<DeclRefExpr>(subImpl))
		   {
		     DeclRefExpr *decl = cast<DeclRefExpr>(subImpl);
			 if (decl->getNameInfo().getAsString() == "free")
			 {
               //retrieve the name of variable
               std::string varName;
               Expr *callArg = call->getArg(0);
               if (isa<ImplicitCastExpr>(callArg))
               {
                 ImplicitCastExpr *implCallArg = cast<ImplicitCastExpr>(callArg); 
                 Expr *subCallArg = implCallArg -> getSubExpr();
                 if (isa<ImplicitCastExpr>(subCallArg))
                 {
                   ImplicitCastExpr *implCastExpr = cast<ImplicitCastExpr>(subCallArg);
				   Expr *subExpr = implCastExpr -> getSubExpr();
                   varName = GetNameOfVariable(subExpr);
                 } 
                 //search for this name in the varNames vector
                 //get index of found element
                 int index = -1;
                 for (int i = 0; i < varNames.size(); i++)
                 {
                   if (varNames.at(i) == varName)
                   { 
                     index = i;
                     break;
                   }
                 } 
                 if (index != -1)
                 {
                   TheRewriter.InsertTextBefore(mallocCalls.at(index)->getLocStart(), ("pool" + std::to_string(poolIndex) + ".")); 
				   TheRewriter.ReplaceText(sizesForMalloc.at(index)->getSourceRange(), "");
				   TheRewriter.InsertTextBefore(forStmts.at(index)->getLocStart(), "boost::pool<> pool" + std::to_string(poolIndex) + "(" + TheRewriter.ConvertToString(sizesForMalloc.at(index)) + ");\n");
				   TheRewriter.InsertTextBefore(decl->getLocStart(), "pool" + std::to_string(poolIndex) + "."); 
				   poolIndex++;
                 }
               }
             }
           else 
           {
           		FunctionDecl *functionDecl = call->getDirectCallee();
				if (functionDecl == 0) break;
				Stmt *functionBody = functionDecl->getBody();
				for (Stmt::child_iterator child_b = functionBody->child_begin(), child_e = functionBody->child_end(); child_b != child_e; ++child_b)
				{
					if (isa<BinaryOperator>(*child_b))
       				{
         				//search for: "varName = (type*)malloc(sizeofsth);"
         				std::cout << "Bin in for\n";
         				BinaryOperator *bin = cast<BinaryOperator>(*child_b);
         				Expr *right =  bin->getRHS(); 
         				Expr *left = bin->getLHS();
         				if (isa<CStyleCastExpr>(right))
		 				{
		   					CStyleCastExpr *cStyleCastExpr = cast<CStyleCastExpr>(right);
		   					Expr *subCast = cStyleCastExpr -> getSubExpr();
  		   					if (isa<CallExpr>(subCast))
           					{
             					CallExpr *call = cast<CallExpr>(subCast);
             					Expr *callee = call->getCallee();
             					if (isa<ImplicitCastExpr>(callee))
			 					{
               						ImplicitCastExpr *impl = cast<ImplicitCastExpr>(callee);
               						Expr *subImpl = impl->getSubExpr();
               						if (isa<DeclRefExpr>(subImpl))
			   						{
				 						DeclRefExpr *decl = cast<DeclRefExpr>(subImpl);
				 						if (decl->getNameInfo().getAsString() == "malloc")
				 						{
				 							std:: cout<< "pool";
				   							TheRewriter.InsertTextBefore(decl->getLocStart(), ("pool" + std::to_string(poolIndex) + ".")); 
				   							TheRewriter.ReplaceText(call->getArg(0)->getSourceRange(), "");
				 						}
			   						}
			 					}	
		   					}
 		 				}  
       				}
       				else if (isa<DeclStmt>(*child_b))
       				{
       					std::cout << "DECL in for\n";
       					DeclStmt *declStmt = cast<DeclStmt>(*child_b);
       					Decl *subDecl = declStmt->getSingleDecl();
       					if (isa<VarDecl>(subDecl))
       					{
       						VarDecl *varDecl = cast<VarDecl>(subDecl);
       						Expr *subVarDecl = varDecl->getInit();
       						if (isa<CStyleCastExpr>(subVarDecl))
       						{
       							CStyleCastExpr *cStyleCastExpr = cast<CStyleCastExpr>(subVarDecl);
       							Expr *subCStyleCastExpr = cStyleCastExpr -> getSubExpr();
       							if (isa<CallExpr>(subCStyleCastExpr))
       							{
       								CallExpr *callExpr = cast<CallExpr>(subCStyleCastExpr);
       								Expr *callee = callExpr->getCallee();
             						if (isa<ImplicitCastExpr>(callee))
			 						{
               							ImplicitCastExpr *impl = cast<ImplicitCastExpr>(callee);
               							Expr *subImpl = impl->getSubExpr();
               							if (isa<DeclRefExpr>(subImpl))
			   							{
				 							DeclRefExpr *decl = cast<DeclRefExpr>(subImpl);
				 							if (decl->getNameInfo().getAsString() == "malloc")
				 							{
				 								std:: cout<< "pool";
				   								TheRewriter.InsertTextBefore(decl->getLocStart(), ("pool" + std::to_string(poolIndex) + ".")); 
				   								TheRewriter.ReplaceText(call->getArg(0)->getSourceRange(), "");
				 							}
				 						}
				 					}
       							}
       						} 
       					}
      				 }
           		}
           	}
           }
         } 
	   }   
    }
    return true;
  }

  std::string GetNameOfVariable(Expr* expr)
  { 
    if (isa<DeclRefExpr>(expr))
    { 
      DeclRefExpr *decl = cast<DeclRefExpr>(expr);
      return decl->getNameInfo().getAsString();
    }
    return "";
  }

private:
  Rewriter &TheRewriter;
};

class MyASTConsumer : public clang::ASTConsumer {
public:
  explicit MyASTConsumer(Rewriter &R)
    : Visitor(R) {}

  virtual bool HandleTopLevelDecl(DeclGroupRef DR) {
    for (DeclGroupRef::iterator b = DR.begin(), e = DR.end(); b != e; ++b)
    {
    	Visitor.TraverseDecl(*b);
    }
    return true; 
  }
private:
  MyVisitor Visitor;
};


int main(int, char **argv)
{
    using clang::CompilerInstance;
    using clang::TargetOptions;
    using clang::TargetInfo;
    using clang::FileEntry;
    using clang::DiagnosticOptions;
    using clang::TextDiagnosticPrinter;
    using clang::SrcMgr::CharacteristicKind;
    using clang::StringRef;
    using clang::DirectoryLookup;
    using llvm::MemoryBuffer;
    using llvm::StringRef;
    using clang::LangOptions;
    using clang::FrontendOptions;
    using clang::LangStandard;
    using clang::CompilerInvocation;
    using clang::InitializePreprocessor;
    using clang::Preprocessor;
    using clang::PreprocessorOptions;
    using clang::HeaderSearch;
    using clang::HeaderSearchOptions;

    CompilerInstance ci;
    DiagnosticOptions diagnosticOptions;
    ci.createDiagnostics();

    CompilerInvocation *invocation = new CompilerInvocation;
    ci.setInvocation(invocation);

    ci.getHeaderSearchOpts().ResourceDir = "/home/evgenia/clang-llvm/build/lib/clang/3.5.0/include/";

    LangOptions &langOpts = ci.getLangOpts();
    langOpts.RTTI = 1;
    langOpts.Bool = 1;
    langOpts.CPlusPlus11 = 1;
    langOpts.GNUKeywords = 1;
    langOpts.CXXExceptions = 1;
    langOpts.POSIXThreads = 1;
    langOpts.SpellChecking = 1;

    ci.getInvocation().setLangDefaults(langOpts, clang::IK_CXX, LangStandard::lang_gnucxx11);

    llvm::IntrusiveRefCntPtr<TargetOptions> pto( new TargetOptions() );
    pto->Triple = llvm::sys::getDefaultTargetTriple();

    llvm::IntrusiveRefCntPtr<TargetInfo> pti(TargetInfo::CreateTargetInfo(ci.getDiagnostics(), pto.getPtr()));
    ci.setTarget(pti.getPtr());

    ci.createFileManager();
    auto &fileManager = ci.getFileManager();

    ci.createSourceManager(fileManager);

    llvm::IntrusiveRefCntPtr<HeaderSearchOptions> headerSearchOpts( new HeaderSearchOptions() );

    ci.createPreprocessor(clang::TU_Complete);
    auto &pp = ci.getPreprocessor();
    ci.getPreprocessorOpts().UsePredefines = false;
    pp.getBuiltinInfo().InitializeBuiltins(pp.getIdentifierTable(), langOpts);

    llvm::IntrusiveRefCntPtr<clang::HeaderSearchOptions> hso( new clang::HeaderSearchOptions());
    HeaderSearch headerSearch(hso,
                              ci.getSourceManager(),
                              ci.getDiagnostics(),
                              ci.getLangOpts(),
                              &ci.getTarget());

    hso->AddPath("/usr/include/", 
                             clang::frontend::Angled, 
                             false, 
                             false);


    hso->AddPath("/usr/lib/gcc/i686-linux-gnu/4.8/include/",
                                clang::frontend::Angled,
                                false, 
                                false);

    hso->AddPath("/usr/include/i386-linux-gnu/",
                                clang::frontend::Angled,
                                false, 
                                false);

    hso->AddPath("/usr/include/i386-linux-gnu/c++/4.8/",
                                clang::frontend::Angled,
                                false, 
                                false);


    hso->AddPath("/usr/include/c++/4.8/",
                                clang::frontend::Angled,
                                false, 
                                false);

    clang::InitializePreprocessor(ci.getPreprocessor(), 
                                  ci.getPreprocessorOpts(),
                                  *hso,
                                  ci.getFrontendOpts());
    ci.createASTContext();
    Rewriter TheRewriter;
    TheRewriter.setSourceMgr(ci.getSourceManager(), ci.getLangOpts());

    const FileEntry *pFile = fileManager.getFile(argv[1]);
    auto &sourceManager = ci.getSourceManager();

    sourceManager.createMainFileID(pFile);
    ci.getDiagnosticClient().BeginSourceFile(ci.getLangOpts(), &pp);

    MyASTConsumer astConsumer(TheRewriter);
    //ci.setASTConsumer(&astConsumer);

    clang::ParseAST(pp, &astConsumer, ci.getASTContext());
    //ci.getDiagnosticClient().EndSourceFile();
    const RewriteBuffer *RewriteBuf = TheRewriter.getRewriteBufferFor(sourceManager.getMainFileID());
    llvm::outs() << string(RewriteBuf->begin(), RewriteBuf->end());

    return 0;
}
